//
// (c) Polar Pyramid                                                                                                             
// suggestions to stolken@kabelfoon.nl and/or stolk@xs4all.nl
//

#ifndef QUAT_INL
#define QUAT_INL

#define EPSILON 0.00000000001


#include <assert.h>


#include "vector3.inl"
#include "quat.h"
#include "matrix44.h"

inline bool IsZero(double f)
{
  return (f < EPSILON && f > -EPSILON);
}


/////////////////////////////////////////////////////////////////////////////
//
// Default Constructor
//

inline Quat::Quat()
  : v(Vector3()), w(1)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Copy Constructor
//

inline Quat::Quat(const Quat& q)
  : v(q.v), w(q.w)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// This constructor is used in operators and sets the member variables 
// directly
//

inline Quat::Quat(float qw, const Vector3 &qv)
  : v(qv), w(qw)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// This constructor is useful for creating rotations by specifying the
// angle of rotation and the axis around which you want to rotate,
// this axis does not have to be normalized
//

inline Quat::Quat(const Vector3 &axis, float angle)
{
  if (!IsZero(angle))
  {
    w = cos(0.5f*angle);
    v = ((float) -sin(0.5f*angle))*axis;
  }
  else
  {
    w = 1.0;
    v = Vector3(0.0f, 0.0f, 0.0f);
  }
}

/////////////////////////////////////////////////////////////////////////////
//
// 4 Element Constructor
//

inline Quat::Quat(float qx, float qy, float qz, float qw)
  : v(qx, qy, qz), w(qw)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Array Access
//

inline float& Quat::operator [](int i)
{
  switch (i)	
  {
  case 0: return v.x;
  case 1: return v.y;
  case 2: return v.z;
  case 3: return w;
  }
  assert(false);
  return w;
}

inline const float& Quat::operator [](int i) const
{
  switch (i)	
  {
  case 0: return v.x;
  case 1: return v.y;
  case 2: return v.z;
  case 3: return w;
  }
  assert(false);
  return w;
}

/////////////////////////////////////////////////////////////////////////////
//
// Real
//

inline float Quat::Real(void) const
{
  return w;
}

/////////////////////////////////////////////////////////////////////////////
//
// Imaginary
//

inline Vector3 Quat::Imaginary(void) const
{
  return v;
}

/////////////////////////////////////////////////////////////////////////////
//
// X
//

inline float Quat::X(void) const
{
  return v.x;
}

/////////////////////////////////////////////////////////////////////////////
//
// Y
//

inline float Quat::Y(void) const
{
  return v.y;
}

/////////////////////////////////////////////////////////////////////////////
//
// Z
//

inline float Quat::Z(void) const
{
  return v.z;
}

/////////////////////////////////////////////////////////////////////////////
//
// Abs
//

inline double Quat::Abs(void) const
{
  return sqrt(Magnitude());
}

/////////////////////////////////////////////////////////////////////////////
//
// Conj
//

inline void Quat::Conj(void)
{
  v = -v;
}

/////////////////////////////////////////////////////////////////////////////
//
// Inverse
//

inline Quat Quat::Inverse(void) const
{
  Quat q = *this;
  q.Conj();
  q.Normalize();
  return q;
}

/////////////////////////////////////////////////////////////////////////////
//
// Normalize
//

inline void Quat::Normalize(void)
{
  float c = Abs();
  assert(!IsZero(c));
  w /= c;
  v /= c;
}

/////////////////////////////////////////////////////////////////////////////
//
// magnitude
//

inline double Quat::Magnitude(void) const
{
  return v.lengthSqr() + w*w;
}

/////////////////////////////////////////////////////////////////////////////
// Math Operators
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
// -Quat
//

inline Quat Quat::operator - () const
{
  return Quat(-w, -v);
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat += Quat
//

inline Quat& Quat::operator += (const Quat& q)
{
  w += q.w;
  v += q.v;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat -= Quat
//

inline Quat& Quat::operator -= (const Quat& q)
{
  w -= q.w;
  v -= q.v;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat *= Quat
//

inline Quat& Quat::operator *= (const Quat& q)
{
  Vector3 mul(w*q.X() + q.w*X() + Y()*q.Z() - Z()*q.Y(),
	      w*q.Y() + q.w*Y() + Z()*q.X() - X()*q.Z(),
	      w*q.Z() + q.w*Z() + X()*q.Y() - Y()*q.X());
  w = w*q.w   - X()*q.X() - Y()*q.Y() - Z()*q.Z();
  v = mul; // do not change the order of these lines! (w needs old v) 
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat /= Quat
//

inline Quat& Quat::operator /=(const Quat& q)
{
  *this *= q.Inverse();
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat ^= Quat
//

inline Quat& Quat::operator ^=(const Quat& q)
{
  *this = Exp(q*Log(*this));
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat *= float
//

inline Quat& Quat::operator *=(float c)
{
  w *= c;
  v *= c;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat /= float
//

inline Quat& Quat::operator /=(float c)
{
  assert(!IsZero(c));
  *this = *this/c;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat * float
//

inline Quat Quat::operator *(float c) const
{
  return Quat(c*w, c*v);
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat / float
//

inline Quat Quat::operator /(float c) const
{
  assert(!IsZero(c));
  return Quat(w/c, v/c);
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat == Quat
//

inline bool Quat::operator == (const Quat& q) const
{
  return ((w == q.w) && (v == q.v));
}

/////////////////////////////////////////////////////////////////////////////
//
// Quat != Quat
//

inline bool Quat::operator != (const Quat& q) const
{
  return (!(*this == q));
}

/////////////////////////////////////////////////////////////////////////////
// Friend functions
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
// float * Quat
//

inline Quat operator *(float c, const Quat& q)
{
  return Quat(c*q.w, c*q.v);
}

/////////////////////////////////////////////////////////////////////////////
//
// Square(Quat)
//

inline Quat Square(const Quat& q)
{
  return Quat(q.w*q.w - q.v.lengthSqr(), 2*q.w*q.v);
}

/////////////////////////////////////////////////////////////////////////////
//
// Exp(Quat)
//

inline Quat Exp(const Quat &q1)
{
  Quat q = q1;

  float s = q.w;
  float se = exp(s);
  q.w = 0.0f;
  float scale = se;
  float theta;
  if ((theta = q.Abs()) > 0.0001f)
    scale *= sin(theta)/theta;
  q.v *= scale;
  q.w = se*cos(theta);
  return q;
}

/////////////////////////////////////////////////////////////////////////////
//
// Log(Quat)
//

inline Quat Log(const Quat &q1)
{  
  Quat q = q1;

  float sl = q.Abs();
  q.Normalize();
  float s = q.w;
  q.w = 0.0f;
  float scale = q.Abs();
  float theta = atan2(scale, s);
  if (scale > 0.0f)
    scale = theta/scale;
  q.v *= scale;
  q.w = log(sl);
  return q;
}

/////////////////////////////////////////////////////////////////////////////
//
// Interpolate(Quat, Quat, t)
//
// spherical interpolation between 2 quaternions
//

inline Quat Interpolate(const Quat &p,
			const Quat &q, 
			double t)
{
  Quat qt, q1;
  double sp, sq;

  if ((p - q).Abs() > (p + q).Abs())
    q1 = -q;
  else
    q1 = q;

  double c = (double) (p.w*q1.w + 
		       p.v[0]*q1.v[0] +
		       p.v[1]*q1.v[1] +
		       p.v[2]*q1.v[2]);

  if (c > -0.9999999999)
  {
    if (c < 0.9999999999)
    {
      double o = acos(c);
      double s = 1.0/sin(o);
      sp = sin((1.0-t)*o)*s;
      sq = sin(t*o)*s;
    }
    else
    {
      sp = 1.0-t;
      sq = t;
    }
    qt.w   = sp*p.w   + sq*q1.w;
    qt.v[0] = sp*p.v[0] + sq*q1.v[0];
    qt.v[1] = sp*p.v[1] + sq*q1.v[1];
    qt.v[2] = sp*p.v[2] + sq*q1.v[2];
  }
  else
  {
    qt.w   =  p.v[2];
    qt.v[0] = -p.v[1];
    qt.v[1] =  p.v[0];
    qt.v[2] = -p.w;
    sp = sin((0.5-t)*M_PI);
    sq = sin(t*M_PI);
    qt.v[0] = sp*p.v[0] + sq*qt.v[0];
    qt.v[1] = sp*p.v[1] + sq*qt.v[1];
    qt.v[2] = sp*p.v[2] + sq*qt.v[2];
  }
  return qt;
}

//
// InsertInMatrix
//

inline void Quat::InsertInMatrix(float m[4][4]) const
{
  float x2 = X() + X();
  float y2 = Y() + Y();
  float z2 = Z() + Z();
  float wx = w*x2;
  float wy = w*y2; 
  float wz = w*z2;
  float xx = X()*x2;
  float xy = X()*y2;
  float xz = X()*z2;
  float yy = Y()*y2;
  float yz = Y()*z2;
  float zz = Z()*z2;
  m[0][0] = 1.0f-yy-zz;
  m[1][0] = xy+wz;
  m[2][0] = xz-wy;
         
  m[0][1] = xy-wz;
  m[1][1] = 1.0f-xx-zz;
  m[2][1] = yz+wx;
         
  m[0][2] = xz+wy;
  m[1][2] = yz-wx;
  m[2][2] = 1.0f-xx-yy;
}

//
// GetXAxis
//  TODO: could be optimized to forget about non used axis values
//

inline Vector3 Quat::GetXAxis(void) const
{
  Matrix44 mat;
  InsertInMatrix(mat.m);
  return mat.GetAxis(0);
}

//
// GetYAxis
//  TODO: could be optimized to forget about non used axis values
//

inline Vector3 Quat::GetYAxis(void) const
{
  Matrix44 mat;
  InsertInMatrix(mat.m);
  return mat.GetAxis(1);
}

//
// GetZAxis
//  TODO: could be optimized to forget about non used axis values
//

inline Vector3 Quat::GetZAxis(void) const
{
  Matrix44 mat;
  InsertInMatrix(mat.m);
  return mat.GetAxis(2);
}

inline void Quat::Dump(void) const
{
  cout << "QUAT with w = " << w << " and vector is: ";
  v.Dump();
}

#endif
