//
// (c) Polar Pyramid                                                                                                             
// suggestions to stolken@kabelfoon.nl and/or stolk@xs4all.nl
//

#ifndef VECTOR4_INL
#define VECTOR4_INL

#include <assert.h>

#include "vector3.h"
#include "vector4.h"

/////////////////////////////////////////////////////////////////////////////
//
// Default Constructor
//

inline Vector4::Vector4()
 : x(0), y(0), z(0), w(0)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Copy Constructor
//

inline Vector4::Vector4(const Vector4& v)
  : x(v.x), y(v.y), z(v.z), w(v.w)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Element Constructor
//

inline Vector4::Vector4(float ix, float iy, float iz, float iw)
  : x(ix), y(iy), z(iz), w(iw)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Array Access
//

inline float& Vector4::operator [](int i)
{
  switch (i)	
  {
  case 0: return x;
  case 1: return y;
  case 2: return z;
  case 3: return w;
  }
  assert(false);
  return x;
}

inline const float& Vector4::operator [](int i) const
{
  assert(i >= 0 && i <= 3);
  switch (i)	
  {
  case 0: return x;
  case 1: return y;
  case 2: return z;
  case 3: return w;
  }
  assert(false);
  return x;
}

/////////////////////////////////////////////////////////////////////////////
//
// Assignment
//

inline Vector4& Vector4::assign(float ix, float iy, float iz, float iw)
{
  x = ix;
  y = iy;
  z = iz;
  w = iw;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
// Math Operators
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
// - Vector4
//

inline Vector4 Vector4::operator -() const
{
  return Vector4(-x, -y, -z, -w);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 += Vector4
//

inline Vector4& Vector4::operator += (const Vector4& v)
{
  x += v.x; 
  y += v.y;
  z += v.z;
  w += v.w;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 -= Vector4
//

inline Vector4& Vector4::operator -= (const Vector4& v)
{
  x -= v.x; 
  y -= v.y;
  z -= v.z;
  w -= v.w;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 += Vector3
//

inline Vector4& Vector4::operator += (const Vector3& v)
{
  x += v.x; 
  y += v.y;
  z += v.z;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 -= Vector3
//

inline Vector4& Vector4::operator -= (const Vector3& v)
{
  x -= v.x; 
  y -= v.y;
  z -= v.z;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 *= float
//


inline Vector4& Vector4::operator *= (float c)
{
  x *= c; 
  y *= c;
  z *= c;
  w *= c;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 /= float
//


inline Vector4& Vector4::operator /= (float c)
{
  x /= c; 
  y /= c;
  z /= c;
  w /= c;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 + Vector4
//


inline Vector4 Vector4::operator + (const Vector4& v) const
{
  return Vector4(x + v.x, 
		 y + v.y,
		 z + v.z,
		 w + v.w);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 - Vector4
//


inline Vector4 Vector4::operator - (const Vector4& v) const
{
  return Vector4(x - v.x, 
		 y - v.y,
		 z - v.z,
		 w - v.w);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 + Vector3
//


inline Vector4 Vector4::operator + (const Vector3& v) const
{
  return Vector4(x + v.x, 
		 y + v.y,
		 z + v.z,
		 w);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 - Vector3
//


inline Vector4 Vector4::operator - (const Vector3& v) const
{
  return Vector4(x - v.x, 
		 y - v.y,
		 z - v.z,
		 w);
}


/////////////////////////////////////////////////////////////////////////////
//
// Vector4 / float
// 


inline Vector4 Vector4::operator / (float c) const
{
  assert(c != 0);
  return Vector4(x/c, 
		 y/c,
		 z/c,
		 w/c);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 == Vector4
//


inline bool Vector4::operator == (const Vector4& v) const
{
  return ((x == v.x) && 
          (y == v.y) &&
          (z == v.z) &&
          (w == v.w));
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 != Vector4
//


inline bool Vector4::operator != (const Vector4& v) const
{
  return (!(*this == v));
}

/////////////////////////////////////////////////////////////////////////////
//
// float * Vector4
//

inline Vector4 operator * (float c, Vector4 const &v)
{
  return Vector4(v.x*c, 
		 v.y*c,
		 v.z*c,
		 v.w*c);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector4 * float
//


inline Vector4 operator * (Vector4 const &v, float c)
{
  return Vector4(v.x*c, 
		 v.y*c,
		 v.z*c,
		 v.w*c);
}

inline void Vector4::Dump(void) const
{
  cout << "Vector4 " << x << "," << y << "," << z <<  "," << w << "\n";
}

/////////////////////////////////////////////////////////////////////////////
//
// length : returns the length of the vector4 as a double, note that 
//          this has to return a double because it uses the sqrt function
//

inline double Vector4::length() const
{
  return sqrt(lengthSqr());
}

/////////////////////////////////////////////////////////////////////////////
//
// lengthSqr : returns the squared length of the vector4
//


inline float Vector4::lengthSqr() const
{
  return x*x + y*y + z*z + w*w;
}

/////////////////////////////////////////////////////////////////////////////
//
// length3d : returns the length of the vector4 as a double, note that 
//          this has to return a double because it uses the sqrt function
//


inline double Vector4::length3d() const
{
  return sqrt(lengthSqr3d());
}

/////////////////////////////////////////////////////////////////////////////
//
// lengthSqr3d : returns the squared 3d length of the vector4
//

inline float Vector4::lengthSqr3d() const
{
  return x*x + y*y + z*z;
}

/////////////////////////////////////////////////////////////////////////////
//
// normalize : will normalize the vector4, make it of length 1
//

inline void Vector4::normalize(void)
{
  double len = length();
  if (len != 0.0)
    *this /= ((float) len);      // avoid SIGFPE
}

//
// ToVector3
//

inline Vector3 Vector4::ToVector3(void) const
{
  return Vector3(x, y, z);
}

/////////////////////////////////////////////////////////////////////////////
//
// dot : returns the dot product of 2 vectors
//

inline float Vector4::dot(const Vector4& v2) const
{
  return x*v2.x + y*v2.y + z*v2.z + w*v2.w;
}

/////////////////////////////////////////////////////////////////////////////
//
// dot3d : returns the dot product of 2 vector4s for 3d only
//

inline float Vector4::dot3d(const Vector4& v2) const
{
  return x*v2.x + y*v2.y + z*v2.z;
}

/////////////////////////////////////////////////////////////////////////////
//
// distanceSqr : return the squared distance between 2 given points which
//               avoids taking an expensive sqrt call. Usefull when comparing
//               different distances.
//

inline float Vector4::distanceSqr(const Vector4& v2)
{
  return
  (x - v2.x)*(x-v2.x) +
	(y - v2.y)*(y-v2.y) +
	(z - v2.z)*(z-v2.z) +
	(w - v2.w)*(w-v2.w);
}

/////////////////////////////////////////////////////////////////////////////
//
// distance : return the distance between 2 given points, note that
//            this has to return a double because it uses the sqrt function
//

inline double Vector4::distance(const Vector4& v2)
{
  return sqrt(distanceSqr(v2));
}

/////////////////////////////////////////////////////////////////////////////
//
// distanceSqr3d : return the squared distance between 2 given points in 3d
//                 only (z is ignored)
//

inline float Vector4::distanceSqr3d(const Vector4& v2)
{
  return
  (x - v2.x)*(x-v2.x) +
	(y - v2.y)*(y-v2.y) +
	(z - v2.z)*(z-v2.z);
}

/////////////////////////////////////////////////////////////////////////////
//
// distance : return the distance between 2 given points, note that
//            this has to return a double because it uses the sqrt function
//

inline double Vector4::distance3d(const Vector4& v2)
{
  return sqrt(distanceSqr3d(v2));
}



#endif


