//
// virtball.cpp
//
// This file demonstrates the use of Vector4 and Matrix44 classes,
// that are provided by the Polar Pyramid Matrix44 library.
// It does this by implementing a virtual ball interface.
//
// In a 3D application, it is important to assign a transformation
// matrix in each object in your world, including cameras and lights.
// This way, you can easily manipulate any element of your scene
// with full 6 degrees of freedom.
//
// BUILD CMDS:
//   g++ -c -Wall -pipe -ggdb -O3 virtball.cpp
//   g++ -o virtball virtball.o -L/usr/local/lib/mesa -lglut -lMesaGLU -lMesaGL -L/usr/X11R6/lib -ltiff -lX11 -lXext -lXmu -lXt -lXi -lSM -lICE
//
// If you have any questions, you can reach me at
// stolk@xs4all.nl
// For questions about matrix44 lib, ask
// stolken@kabelfoon.nl
//
//   Bram Stolk, 990408
//
// 990417 Bram
// Added specular hilight, and lamp-orbit.
// this revealed a bug, which maybe in Mesa.
// After some orbitting, the spec hilight is messed up,
// when using mesa-3.1b1 from CVS built for linux-alpha-elf.
// When built for linux-386-glide, it's OK. When using
// mesa 2.6 or 3.0, specular is messed up from the very start.
// Also, try export LOCAL_VIEWER=1, for me, the specular disappears.
// If you experience similar problems, please mail me.
//

#define GLUT_DISABLE_ATEXIT_HACK

#include <stdlib.h>
#include <stdio.h>

#include <vector3.inl>
#include <vector4.inl>
#include <matrix44.inl>
#include <quat.inl>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>




class WorldObject
{
  public:

    Matrix44 trf;
    Vector4 colour;
    WorldObject() { colour = Vector4(1,1,1,1); }
    virtual void Render(void) {}
};


class Ball : public WorldObject
{
  public:
  Ball()
  {
    obj = gluNewQuadric();
    gluQuadricDrawStyle(obj,(GLenum)GLU_FILL);
    gluQuadricNormals(obj,(GLenum)GLU_SMOOTH);
    gluQuadricTexture(obj,GL_FALSE);
  }
  virtual void Render(void) 
  {
    float col[4] = { colour.x, colour.y, colour.z, colour.w };
    glMaterialfv(GL_FRONT, GL_DIFFUSE,   col); 
#if 1
    float spc[4] = { 1,1,1,1 };
    glMaterialfv(GL_FRONT, GL_SPECULAR,  spc);
    float pow[1] = { 8 };
    glMaterialfv(GL_FRONT, GL_SHININESS, pow);
#endif
    gluSphere(obj,2.0,16,16);
  }
  static GLUquadricObj *obj;
};


GLUquadricObj *Ball::obj=0;


class Camera : public WorldObject
{
};


class Lamp : public WorldObject
{
  public:
  Vector3 coi;
  Vector4 colour;
  // NOTE: Mesa-3.1b1 is bugged. Only Mesa-3.0 will do 
  // spotlights properly.
  bool is_spot;
  int lnr;
  static int avlight; // available light nr
  Lamp()
  {
    lnr = GL_LIGHT0 + avlight++;
    is_spot = false;
  }
  virtual void Render(void) 
  {
    Vector3 v = trf.GetPos();
    Vector3 d = coi - v;

    float position[] ={ v.x, v.y, v.z, 1.0 };
    float position2[]={ v.x, v.y, v.z, 0.0 };
    float dir[]      ={ d.x, d.y, d.z};
    float col[]      ={ colour.x, colour.y, colour.z, 1.0 };
 
    if (is_spot)
    {
      glLightf( (GLenum)lnr, GL_SPOT_CUTOFF,    15.0);
      glLightfv((GLenum)lnr, GL_SPOT_DIRECTION, dir);
      glLightf( (GLenum)lnr, GL_SPOT_EXPONENT,  20.0);
      glLightfv((GLenum)lnr, GL_POSITION,       position);
    }
    else
    {
      glLightfv((GLenum)lnr, GL_POSITION,       position2);
      glLightfv((GLenum)lnr, GL_DIFFUSE,        col);
    }
    glEnable((GLenum)lnr);
  }
};

int Lamp::avlight = 0;


// Some objects we use in our scene.
Lamp   ourlamp;
Ball   ourball1;
Ball   ourball2;
Ball   ourball3;
Camera ourcam;


// For virtual ball interface
static int dragmode=0;
static int orgx=0;
static int orgy=0;



// Let the camera orbit the centre of interest.
static void camorbit(float x, float y)
{
  Matrix44 orbtrf = ourcam.trf.Orbit(x,y);
  Matrix44 newtrf = orbtrf * ourcam.trf;
  if (!newtrf.IsValid())
    cout << "Error - invalid newtrf for cam\n";
  ourcam.trf = newtrf;
}

// Let the light orbit the centre of interest.
static void lmporbit(float x, float y)
{
  Matrix44 orbtrf = ourcam.trf.Orbit(x,y);
  Matrix44 newtrf = orbtrf * ourlamp.trf;
  if (!newtrf.IsValid())
    cout << "Error - invalid newtrf for lamp\n";
  ourlamp.trf = newtrf;
}


// Call back for mouse motion
static void motion(int x, int y)
{
  if (!dragmode) return;
  int dx = x - orgx;
  int dy = y - orgy;
  if (!dx && !dy) return;
  if (dragmode == 1)
    camorbit(-dx,dy);
  if (dragmode == 2)
    lmporbit(-dx,dy);
  orgx = x;
  orgy = y;
}


// Create worldobjects.
static void SetupWorld()
{
  ourlamp.trf.Unit();
  ourlamp.trf.SetPos(Vector3(0,30,-30));
  ourlamp.coi = Vector3(0,0,0);
  ourlamp.colour = Vector4(1.0, 1.0 ,1.0, 1);

  ourball1.trf.Unit();
  ourball1.trf.SetPos(Vector3(-5,0,0));
  ourball1.colour = Vector4(1.0, 0.5 ,0.5, 1);

  ourball2.trf.Unit();
  ourball2.trf.SetPos(Vector3(0,0,0));
  ourball2.colour = Vector4(0.5, 1.0 ,0.5, 1);

  ourball3.trf.Unit();
  ourball3.trf.SetPos(Vector3(5,0,0));
  ourball3.colour = Vector4(0.5, 0.5 ,1.0, 1);

  Matrix44 camtrf(Quat(0,0,0,1));
  camtrf.SetPos(Vector3(0,0,40));
  ourcam.trf = camtrf;
}


// Callback for mouseclick
static void mouse(int button, int state, int x, int y)
{
  if (state == GLUT_UP)
  {
    dragmode = 0;
    return;
  }

  if (button == GLUT_LEFT_BUTTON)
  {
    dragmode = 1;
    orgx = x;
    orgy = y;
    return;
  }

  if (button == GLUT_RIGHT_BUTTON)
  {
    dragmode = 2;
    orgx = x;
    orgy = y;
    return;
  }
 
  if (button == GLUT_MIDDLE_BUTTON)
  {
    SetupWorld();
    return;
  }
}


// Setup projection matrix
static void SetupProjection(void)
{
  glMatrixMode(GL_PROJECTION); 
  glLoadIdentity();                                                           
  glFrustum(-1.0,1.0,-1.0,1.0,5.0,600.0);
}


// Print info about gl implementation
static void About(void)
{
  fprintf(stderr,"GL_VERSION: %s\n", (char *) glGetString(GL_VERSION));
  fprintf(stderr,"GL_RENDERER: %s\n", (char *) glGetString(GL_RENDERER));
  fprintf(stderr,"GL_VENDOR: %s\n", (char *) glGetString(GL_VENDOR));
  fprintf(stderr,"GLUT_API_VERSION: %d\n", GLUT_API_VERSION);
}


// Display a frame
void Display(void)
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glClearColor(0,0,1,1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Our entire world database
  WorldObject *objlist[] = 
  {
    &ourlamp, &ourball1, &ourball2, &ourball3, 0
  };

  if (!ourcam.trf.IsValid())
  {
    cout << "Error - invalid cam trf matrix:\n";
    ourcam.trf.Dump();
  }

  Matrix44 view_trf = ourcam.trf;
  if (view_trf.Invert()==0)
    fprintf(stderr,"Error - matrix invert failed\n");

  int i=0;
  glMatrixMode(GL_MODELVIEW);
  while (objlist[i])
  {
    glLoadMatrixf((float *) view_trf.m);
    glMultMatrixf((float *) objlist[i]->trf.m);
    objlist[i]->Render();
    i++;
  }

  // Make glut swap its buffers
  glutSwapBuffers();

  GLenum error = glGetError();
  while (error != GL_NO_ERROR)
  {
    fprintf(stderr, "OpenGL error: %s\n", gluErrorString(error));
    error = glGetError();
  }
}


void Idle(void)
{
  glutPostRedisplay();
}


// callback for keypresses
void KeyboardDown(unsigned char key, int x, int y)
{
  static float speed = 10.0;
  switch (key)
  {
    case 27:
      _exit(1);
      break;

    case 'h' :
      camorbit(-speed,0);
      break;
    case 'l' :
      camorbit(speed,0);
      break;
    case 'k' :
      camorbit(0,-speed);
      break;
    case 'j' :
      camorbit(0,speed);
      break;
  }
}


// callback for reshape events
void Reshape(int w, int h)
{
  glViewport(0, 0, w, h);
}


// initialize ogl
void SetupOGL(int width, int height)
{
  glutInitWindowSize(width, height);
  glutInitWindowPosition(0,0);
  int argc=2;
  char *argv[]={"virtball","-gldebug",0};
  glutInit(&argc,argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  int retval = glutCreateWindow("virtball");
  if (!retval)
    fprintf(stderr,"glutCreateWindow failed\n");

  // Set the callbacks for glut
  glutReshapeFunc(Reshape);
  glutDisplayFunc(Display);
  glutKeyboardFunc(KeyboardDown);
  glutMotionFunc(motion);
  glutMouseFunc(mouse);
  glutIdleFunc(Idle);

  // Configure our culling
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);
  glFrontFace(GL_CW);
 
  if (getenv("LOCAL_VIEWER"))
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  else
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  // reset texture matrix 
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  // reset modelview matrix
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glEnable(GL_NORMALIZE); // needed to support scaling trfs
  glEnable(GL_RESCALE_NORMAL);

  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_LIGHTING);
  glShadeModel(GL_SMOOTH);

}


// Our main program.
int main(int argc, char *argv[])
{
  SetupOGL(640,480);
  SetupProjection();
  SetupWorld();

  About();
  
  printf("Use leftmouse button to drag camera orbit\n");
  printf("Use rightmouse button to drag lamp orbit\n");
  printf("Use middle button to reset all transformations\n");
  printf("Press ESC to quit\n");

  glutMainLoop();
}

